#include <windows.h>
#include <stdio.h>
#include <malloc.h>

// the next three includes access the USB HID interface
#include <setupapi.h>					
extern "C" {
#include </98DDK/src/hid/inc/hidsdi.h>	/* note: part of Windows DDK */
#include </98DDK/src/hid/inc/hidpi.h>	/* note: also part of Windows DDK */
}

#include "DMXhid.h"

/* global constants */

// vendor ID (Durand Interstellar, Inc.)
#define VID 0x0F48
// product ID
#define PID 0x4201

bool DMX_adapter_found = false;

static int outputReportLength;
static char * outbuffer;
static char * inBuffer;
static HIDP_CAPS capabilities;

HANDLE hidHandle;

// locally used function prototypes
DWORD WINAPI Read_Thread( LPVOID lpParam );
BOOL bOpenHidDevice(HANDLE *HidDevHandle, USHORT vid, USHORT pid);
int ReadFromConverter(void);


/**********************************************************
*
*	Function: bOpenHidDevice
*	Purpose: tries to open a HID device based on VID and PID
*	Parameters: vid - HID device's vendor ID
*				pid - HID device's product ID
*				HidDevHandle - pointer to a handle to the HID device
*	Returns: TRUE, if device is found
*			 FALSE, if device is not found
*
*   Based on the book "USB Complete, Second Edition" by Jan Axelson
*
**********************************************************/

BOOL bOpenHidDevice(HANDLE *HidDevHandle, USHORT vid, USHORT pid)
{
	static GUID HidGuid;						// HID Globally Unique ID: windows supplies us with this value
	HDEVINFO HidDevInfo;						// handle to structure containing all attached HID Device information
	SP_DEVICE_INTERFACE_DATA devInfoData;		// Information structure for HID devices
	BOOLEAN Result;								// result of getting next device information structure
	DWORD Index;								// index of HidDevInfo array entry
	DWORD DataSize;								// size of the DeviceInterfaceDetail structure
	BOOLEAN GotRequiredSize;					// 1-shot got device info data structure size flag
	PSP_DEVICE_INTERFACE_DETAIL_DATA detailData = NULL;// device info data
	DWORD RequiredSize;							// size of device info data structure
	BOOLEAN DIDResult;							// get device info data result
	HIDD_ATTRIBUTES HIDAttrib;					// HID device attributes
				
	// initialize variables
	GotRequiredSize = FALSE;

	// 1) Get the HID Globally Unique ID from the OS
	HidD_GetHidGuid(&HidGuid);

	// 2) Get an array of structures containing information about
	// all attached and enumerated HIDs
	HidDevInfo = SetupDiGetClassDevs(&HidGuid, 
									 NULL,
									 NULL,
									 DIGCF_PRESENT | DIGCF_INTERFACEDEVICE);

	// 3) Step through the attached device list 1 by 1 and examine
	// each of the attached devices.  When there are no more entries in
	// the array of structures, the function will return FALSE.
	Index = 0;									// init to first index of array
	devInfoData.cbSize = sizeof(devInfoData);	// set to the size of the structure
												//    that will contain the device info data
	
	do {
		// Get information about the HID device with the 'Index' array entry
		Result = SetupDiEnumDeviceInterfaces(HidDevInfo, 
											 0,
											 &HidGuid,
											 Index,
											 &devInfoData);
		
		// If we run into this condition, then there are no more entries
		// to examine, we might as well return FALSE at point
		if(Result == FALSE)
		{
			// free the memory allocated for DetailData
			if(detailData != NULL)
				free(detailData);
			
			// free HID device info list resources
			SetupDiDestroyDeviceInfoList(HidDevInfo);
			
			return FALSE;
		}

		if(GotRequiredSize == FALSE)
		{
			// 3) Get the size of the DEVICE_INTERFACE_DETAIL_DATA
			// structure.  The first call will return an error condition, 
			// but we'll get the size of the strucure
			DIDResult = SetupDiGetDeviceInterfaceDetail(HidDevInfo,
														&devInfoData,
														NULL,
														0,
														&DataSize,
														NULL);
			GotRequiredSize = TRUE;

			// allocate memory for the HidDevInfo structure
			detailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(DataSize);
			
			// set the size parameter of the structure
			detailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
		}
			
		// 4) Now call the function with the correct size parameter.  This 
		// function will return data from one of the array members that 
		// Step #2 pointed to.  This way we can start to identify the
		// attributes of particular HID devices.
		DIDResult = SetupDiGetDeviceInterfaceDetail(HidDevInfo,
													&devInfoData,
													detailData,
													DataSize,
													&RequiredSize,
													NULL);

		// 5) Open a file handle to the device.
		*HidDevHandle = CreateFile( detailData->DevicePath,
									GENERIC_READ | GENERIC_WRITE,
									FILE_SHARE_READ | FILE_SHARE_WRITE,
									NULL,
									OPEN_EXISTING,
									0,
									NULL);
			
		// 6) Get the Device VID & PID to see if it's the device we want
		if(*HidDevHandle != INVALID_HANDLE_VALUE)
		{
			HIDAttrib.Size = sizeof(HIDAttrib);
			HidD_GetAttributes(	*HidDevHandle, &HIDAttrib);

			if((HIDAttrib.VendorID == VID) && (HIDAttrib.ProductID == PID))
            {
				// free the memory allocated for DetailData
				if(detailData != NULL)
					free(detailData);
				
				// free HID device info list resources
				SetupDiDestroyDeviceInfoList(HidDevInfo);

				return TRUE;	// found HID device
			}
			
			// 7) Close the Device Handle because we didn't find the device
			// with the correct VID and PID
			CloseHandle(*HidDevHandle);
		}

		Index++;	// increment the array index to search the next entry

	} while(Result == TRUE);

	// free the memory allocated for DetailData
	if(detailData != NULL)
		free(detailData);

	// free HID device info list resources
	SetupDiDestroyDeviceInfoList(HidDevInfo);

	return FALSE;
}


void OpenDmxConverter(void)
{
    PHIDP_PREPARSED_DATA hidParsedData;

	// attempt to open the DMX converter (identified by its IDs)
    if (bOpenHidDevice(&hidHandle, VID, PID))
	{
        DMX_adapter_found = true;

		// get a handle to a buffer that describes the devices capabilities
		HidD_GetPreparsedData(hidHandle, &hidParsedData);

		// extract the capabilities info
		HidP_GetCaps(hidParsedData, &capabilities);
            // report sizes are found in
            // capabilities.InputReportByteLength and capabilities.OutputReportByteLength

		outputReportLength = capabilities.OutputReportByteLength;

		// free the memory allocated when getting the preparsed data
		HidD_FreePreparsedData(hidParsedData);

        // create some buffers
        outbuffer = (char *)malloc(capabilities.OutputReportByteLength + 1);
	    inBuffer = (char *)malloc(capabilities.InputReportByteLength + 1);

        // request info from the DMX Converter
		SendCommand("I");
    }
}


int ReadFromConverter(void)
{
    BOOL readStat;
    DWORD numCharsRead;

	do
	{
		readStat = ReadFile(hidHandle,
							inBuffer,
							capabilities.InputReportByteLength,
							&numCharsRead,
							NULL);
		if (readStat)
		{
			// check for "end of file"
			if (numCharsRead <= 0)
			{
				// must be "end of file"
				return -1;
			}
			else
			{ //  a good read so just print it

				// NOTE: inBuffer[0] is the USB packet ID
				//       inBuffer[1] contains the transmission flags
				//       inBuffer[2] is the length of data to follow
				//       inBuffer[3..64] is the data

				// first terminate the string
				inBuffer[inBuffer[2]+3] = '\0';
				printf("%s", &inBuffer[3]);
			}
		}
		else
		{
			// the read had an error
			MessageBox(NULL, "USB Comm error", "Error", MB_OK);
			// report the error to the caller
			return 1;
		}

	  // check for prompt
	} while (strstr(&inBuffer[3], "\r\n>") == NULL);
	
	// tell the caller it was a good read
	return 0;
}


bool SendCommand(char * command)
{
    BOOL result;
    DWORD numBytesWritten;

    if (DMX_adapter_found)
	{
		// clear the output buffer
		memset(outbuffer, 0, outputReportLength);

		outbuffer[0] = 0;		// USB report ID
		outbuffer[1] = 0x30;	// DMXC flags (RTS=true,DTR=true)

		outbuffer[2] = strlen(command) + 1;		    // DMXC command length (data + return)
		sprintf((char *)&outbuffer[3], "%s\r", command);    // the command and return

		result = WriteFile(hidHandle,
						   &outbuffer[0],
						   outputReportLength,
						   &numBytesWritten,
						   NULL);
		if (!result)
		{
			return false;	// return an error indication
		}

		// read the response from the DMX converter
		ReadFromConverter();
	}

	return true;
}
